#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <cbm.h>
#include <peekpoke.h>
#include <string.h>
#include <6502.h>

#define p_clr  0x93 /* Schoon scherm */
#define p_hlb  0xb0 /* Hoek linksboven */
#define p_hrb  0xae /* Hoek rechtsboven */
#define p_hlo  0xad /* Hoek linksonder */
#define p_hro  0xbd /* Hoek rechtsonder */
#define p_hor  0x60 /* Steep horizontaal */
#define p_ver  0x7d /* Streep verticaal */

#define tool_start   0x1000
#define tool_size    PEEKW(0x47fe)

#define SET_ST(x) { POKE (0x90,x); }
#define GET_ST() (PEEK(0x90))

#define TI_HI 160
#define TI_MID 161
#define TI_LO 162

#define FC3_BANK(x) { POKE (0xdfff,x); }

char		started_from_basic;
char		fc3_detected;
char		filename[32];
char		*filename_extra;
char		secprot[8];
char		tmpstr[256];
#pragma bss-name (push,"ZEROPAGE")
#pragma data-name(push,"ZEROPAGE")
char		fast1541;
char		device,olddevice;
char		flashidx,flashsecsz,flashseccnt;
unsigned	seca,secz;
#pragma bss-name (pop)
#pragma data-name(pop)

#define flashidx_last_12v 6
#define FLASHIDX_IS_12V(x) ( x <= flashidx_last_12v )

/* Flash memories with sector protect should come before
   ones without. */

/*12V 256KB flash roms */
#define flash_idx_CAT28F020		0
#define flash_idx_i28F020		1
#define flash_idx_IS28F020		2
#define flash_idx_AM28F020		3
#define flash_idx_M28F201		4
#define flash_idx_MX28F002T		5
#define flash_idx_MX28F002B		6

/*5V 256KB flash roms */
#define flash_idx_AM29F002T		7
#define flash_idx_AM29F002B		8
#define flash_idx_A29002T		9
#define flash_idx_A29002U		10
#define flash_idx_EN29F002T		11
#define flash_idx_EN29F002B		12
#define flash_idx_HY29F002T		13
#define flash_idx_HY29F002B		14
#define flash_idx_M29F002T		15
#define flash_idx_M29F002B		16
#define flash_idx_MBM29F002T	17
#define flash_idx_MBM29F002B	18
#define flash_idx_MX29F002T		19
#define flash_idx_MX29F002B		20
#define flash_idx_SST39SF020	21
#define flash_idx_SST39VF020	22
#define flash_idx_W39L020		23
#define flash_idx_W49F002B		24
#define flash_idx_W49F002U		25
#define flash_idx_AT49F002      26
#define flash_idx_AT49F002T     27
#define flash_idx_AT49F020      28

/* 5V 512KB flash roms */
#define flash_idx_AM29F040		29
#define flash_idx_A29040B		30
#define flash_idx_EN29F040		31
#define flash_idx_HY29F040		32
#define flash_idx_M29F040		33
#define flash_idx_MBM29F040		34
#define flash_idx_MX29F040		35
#define flash_idx_SST39SF040	36
#define flash_idx_SST39VF040	37
#define flash_idx_W39L040		38
#define flash_idx_AT49F040      39

/* Highest idx that has sector protect */

#define flash_chips_count 40

unsigned	flashrom_ids[] = {
	/* 256KB 12V flashroms */
	0x31bd, /* CAT28F020 */
	0x89bd, /* i28F020 (TMS28F020) */
	0xd5bd, /* IS28F020 */
	0x012a, /* AM28F020 */
	0x20f4, /* M28F201 */
    0xc22d, /* MX28F002T */
    0xc22e, /* MX28F002B */
	
	/* 256KB 5V flashroms */
	0x01b0, /* AM29F002T */
	0x0134, /* AM29F002B */
	0x378c, /* A29002T */
	0x370d, /* A29002U */
	0x1c92, /* EN29F002T */
	0x1c94, /* EN29F002B */
	0xadb0, /* HY29F002T */
	0xad34, /* HY29F002B */
	0x20b0, /* M29F002T (ST) */
	0x2034, /* M29F002B (ST) */
	0x04b0, /* MBM29F002T */
	0x0434, /* MBM29F002B */
	0xc2b0, /* MX29F002T */
	0xc234, /* MX29F002B */
	0xbfb6, /* SST39SF020 */
	0xbfd6, /* SST39VF020 */
	0xdab5, /* W29L020 */
	0xda25, /* W49F002B */
	0xda0b, /* W49F002U */
	0x1f07, /* AT49F002 */
	0x1f08, /* AT49F002T */
	0x1f0b, /* AT49F020 */

	/* 512KB 5V flashroms */
	0x01a4, /* AM29F040 (TMS29F040, AS29F040) */
	0x3786, /* A29040B (AS29CF040) */
	0x1c04, /* EN29F040 */
	0xad40, /* HY29F040, BM29F040 */
	0x20e2, /* M29F040 */
	0x04a4, /* MBM29F040 */
	0xc2a4, /* MX29F040C */
	0xbfb7, /* SST39SF040 */
	0xbfd7, /* SST39VF040 */
	0xdad6, /* W39L040 */
	0x1f13  /* AT49F040 */
};

/*

   20e3 M29W040
   52a4 AS29F040
   014f AM29LV040
   3792 A29L040
   1c4f EN29LV040
   9d4e PM39F040
   bf24 SST29SF020
   bf13 SST29SF040
   da8c W49F020
   da45 W29C020
*/

const char *flashrom_names[] = {
	/* 256KB 12V flashroms */
	"CAT28F020",
	"i28F020",
	"IS28F020",
	"AM28F020",
	"M28F201",
	"MX28F002T",
	"MX28F002B",

	/* 256KB 5V flashroms */
	"AM29F002T",
	"AM29F002B",
	"A29002T",
	"A29002U",
	"EN29F002T",
	"EN29F002B",
	"HY29F002T",
	"HY29F002B",
	"M29F002T",
	"M29F002B",
	"MBM29F002T",
	"MBM29F002B",
	"MX29F002T",
	"MX29F002B",
	"SST39SF020",
	"SST39VF020",
	"W39L020",
	"W49F002B",
	"W49F002U",
	"AT49F002",
	"AT49F002T",
	"AT49F020",

	/* 512KB 5V flashroms */
	"AM29F040",
	"A29040B",
	"EN29F040",
	"HY29F040",
	"M29F040",
	"MBM29F040",
	"MX29F040",
	"SST39SF040",
	"SST39VF040",
	"W39L040",
	"AT49F040"
};

/* 1 << secsz == sector size */
const char secsz[] = {
	/* 256KB 12V flashroms */
	18, /* CAT28F020 */
	18, /* i28F020 */
	18, /* IS28F020 */
	18, /* AM28F020 */
	18, /* M28F201 */
    17, /* MX28F002T */
    17, /* MX28F002B */

	/* 256KB 5V flashroms */
	16, /* AM29F002T */
	16, /* AM29F002B */
	16, /* A29002T */
	16, /* A29002U */
	16, /* EN29F002T */
	16, /* EN29F002B */
	16, /* HY29F002T */
	16, /* HY29F002B */
	16, /* M29F002T (ST) */
	16, /* M29F002B (ST) */
	16, /* MBM29F002T */
	16, /* MBM29F002B */
	16, /* MX29F002T */
	16, /* MX29F002B */
	12, /* SST39SF020 */
	12, /* SST39VF020 */
	16, /* W39L020 */
	17, /* W49F002B */
	17, /* W49F002U */
	17, /* AT49F002 */
	17, /* AT49F002T */
	18, /* AT49F020 */

	
	/* 512KB 5V flashroms */
	16, /* AM29F040 */
	16, /* A29C040B */
	16, /* EN29F040 */
	16, /* HY29F040 */
	16, /* M29F040 */
	16, /* MBM29F040 */
	16, /* MX29F040C */
	12, /* SST39SF040 */
	12, /* SST39VF040 */
	16, /* W39L040 */
	18  /* AT49F040 */
};

const unsigned seccnt[] = {
	/* 256KB 12V flashroms */
	1, /* CAT28F020 */
	1, /* i28F020 */
	1, /* IS28F020 */
	1, /* AM28F020 */
	1, /* M28F201 */
    2, /* MX28F002T */
    2, /* MX28F002B */


	/* 256KB 5V flashroms */
	4, /* AM29F002T */
	4, /* AM29F002B */
	4, /* A29002T */
	4, /* A29002U */
	4, /* EN29F002T */
	4, /* EN29F002B */
	4, /* HY29F002T */
	4, /* HY29F002B */
	4, /* M29F002T (ST) */
	4, /* M29F002B (ST) */
	4, /* MBM29F002T */
	4, /* MBM29F002B */
	4, /* MX29F002T */
	4, /* MX29F002B */
	64, /* SST39SF020 */
	64, /* SST39VF020 */
	4, /* W39L020 */
	2, /* W49F002B */
	2, /* W49F002U */
	2, /* AT49F002 */
	2, /* AT49F002T */
	1, /* AT49F020 */

	/* 512KB 5V flashroms
	   In a 512KB flash rom the cartridge hardware can only address 256KB,
	   therefore the program will consider them to be 256KB and the sector
	   count is therefore for 256KB. */
	4, /* AM29F040 */
	4, /* A29040B */
	4, /* EN29F040A */
	4, /* HY29F040 */
	4, /* M29F040 */
	4, /* MBM29F040 */
	4, /* MX29F040C */
	64, /* SST39SF040 */
	64, /* SST39VF040 */
	4,  /* W39L040 */
	1   /* AT49F040 */
};

#define flash_topboot		1
#define flash_bottomboot	2
#define flash_sectorprotect	4
#define flash_12v_unprotect	8

const unsigned flash_features[] = {
	/* 256KB 12V flashroms */
	0, /* CAT28F020 */
	0, /* i28F020 */
	0, /* IS28F020 */
	0, /* AM28F020 */
	0, /* M28F201 */
    0, /* MX28F002T */
    0, /* MX28F002B */


	/* 256KB 5V flashroms */
	flash_topboot|flash_sectorprotect|flash_12v_unprotect,		/* AM29F002T */
	flash_bottomboot|flash_sectorprotect|flash_12v_unprotect,	/* AM29F002B */
	flash_topboot|flash_sectorprotect|flash_12v_unprotect,		/* A29002T */
	flash_bottomboot|flash_sectorprotect|flash_12v_unprotect,	/* A29002U */
	flash_topboot|flash_sectorprotect|flash_12v_unprotect,		/* EN29F002T */
	flash_bottomboot|flash_sectorprotect|flash_12v_unprotect,	/* EN29F002B */
	flash_topboot|flash_sectorprotect|flash_12v_unprotect,		/* HY29F002T */
	flash_bottomboot|flash_sectorprotect|flash_12v_unprotect,	/* HY29F002B */
	flash_topboot|flash_sectorprotect|flash_12v_unprotect,		/* M29F002T (ST) */
	flash_bottomboot|flash_sectorprotect|flash_12v_unprotect,	/* M29F002B (ST) */
	flash_topboot|flash_sectorprotect|flash_12v_unprotect,		/* MBM29F002T */
	flash_bottomboot|flash_sectorprotect|flash_12v_unprotect,	/* MBM29F002B */
	flash_topboot|flash_sectorprotect|flash_12v_unprotect,		/* MX29F002T */
	flash_bottomboot|flash_sectorprotect|flash_12v_unprotect,	/* MX29F002B */
	0, /* SST39SF020 */
	0, /* SST39VF020 */
	0, /* W39L020 */
	flash_topboot|flash_sectorprotect|flash_12v_unprotect,		/* W49F002B */
	flash_bottomboot|flash_sectorprotect|flash_12v_unprotect,	/* W49F002U */
	flash_topboot|flash_sectorprotect|flash_12v_unprotect,		/* AT49F002 */
	flash_bottomboot|flash_sectorprotect|flash_12v_unprotect,	/* AT49F002T */
	0,															/* AT49F020 */

	/* 512KB 5V flashroms
	   In a 512KB flash rom the cartridge hardware can only address 256KB,
	   therefore the program will consider them to be 256KB and the sector
	   count is therefore for 256KB. */
	flash_sectorprotect, /* AM29F040 */
	flash_sectorprotect, /* A29040B */
	flash_sectorprotect, /* EN29F040A */
	flash_sectorprotect, /* HY29F040 */
	flash_sectorprotect, /* M29F040 */
	flash_sectorprotect, /* MBM29F040 */
	0, /* MX29F040C */
	0, /* SST39SF040 */
	0, /* SST39VF040 */
	0, /* W39L040 */
	0 /* AT49F040 */
};

extern void install_nmi();
extern char test_nmi();
extern void install_lowcode();
extern char test_ultimax();
extern void flash_magic_unlock();
extern void flash_poke(char data);
extern void flash_enable_12v();
extern void flash_disable_12v();
extern char* pokeaddr;
extern char pokebank;

#define FLASH_POKE(bank,addr,data) { \
  pokebank=0x50|bank; \
  pokeaddr=(char *) (addr|(addr & 0x2000)<<1); \
  flash_poke(data);}


extern void strout(register const char *s);
/*void strout(register const char *s) {
  while (*s!=0) {
    cbm_k_bsout(*s);
    ++s;
  }
}*/

void numout(unsigned v) {
  utoa(v,tmpstr,10);
  strout(tmpstr);
}

extern void serialout(register const char *s);
/*
void serialout(register const char *s) {
  while (*s!=0) {
    cbm_k_ciout(*s);
    ++s;
  }
}*/

void serial_skip(char n);
/*void serial_skip(char n) {
  do {
    cbm_k_acptr();
    n--;
  } while (n!=0);
}*/

void fast1541_disable_screen() {
  if (fast1541) {
    POKE(0xd011,PEEK(0xd011) & 0xef);
    /* The VIC-II decides at raster line 48 wether or not to generate bad
       lines and no bad lines occur at lines before*/
    SEI();
    do {} while (PEEK(0xd012)|(PEEK(0xd011) & 0x80)<=48);
    CLI();
  }
}

void fast1541_enable_screen() {
  if (fast1541)
    POKE(0xd011,PEEK(0xd011) | 0x10);
}

char send_drive_command(const char *s) {
  static char c;
  SET_ST(0);
  cbm_k_listen(device);
  cbm_k_second(0x6f);
  if (c=GET_ST()) {
    cbm_k_unlsn();
    return c;
  }
  serialout(s);
  cbm_k_unlsn();
  return GET_ST();
}

void print_drive_status() {
  /* Device MUST be present, checking st doesn't help after TALK */
  register char *s;

  SET_ST(0);
  fast1541_disable_screen();
  cbm_k_talk(device);
  cbm_k_tksa(0x6f);
  s=tmpstr;
  do {
    *s=cbm_k_acptr();
  } while (*s++!=13);
  *s=0;
  cbm_k_untlk();
  fast1541_enable_screen();
  strout(tmpstr);
}

void handle_st(const char st) {
  if (st & 128)
    strout("Device not present");
  if (st & 64)
    strout("End of file");
  if (st & 16)
    strout("Verify");
  if (st & 2)
    strout("Read");
  if (st & 1)
    strout("Write ");
  if (st)
    strout(" error.\n");
}

void file_open(const char *s, const char chan) {
  cbm_k_listen(device);
  cbm_k_second(0xf0 | chan);
  serialout(s);
  cbm_k_unlsn();
}

void file_close(const char chan) {
 /* Tell device we want to close the directory. */
  cbm_k_listen(device);
  cbm_k_second(0xe0 | chan);
  cbm_k_unlsn();
}

void directory(const char *s) {
  static unsigned size;
  static char c;

  file_open(s,0);  /* Chan 0 = load PRG */
 
  SET_ST(0);
  /* Make device talk the directory. */
  fast1541_disable_screen();
  cbm_k_talk(device);
  cbm_k_tksa(0x60);  /* Chan 0 = load PRG */
  /* Skip load address */
  serial_skip(2);
  do {
    /* Skip BASIC line pointer */
    serial_skip(2);
    if (GET_ST())
      break;
    /* Compute the file size in blocks and print it. */
    size = cbm_k_acptr() | (cbm_k_acptr() << 8);
    if (GET_ST())
      break;
    numout(size);
    cbm_k_bsout(' ');

    /* Get the rest of the directory entry and print it */
    do {
      c = cbm_k_acptr();
      if (c==0 || GET_ST())
        break;
      cbm_k_bsout(c);
    } while (1);
    if (GET_ST())
      break;
    cbm_k_bsout(13);
  } while (1);
  cbm_k_untlk();
  fast1541_enable_screen();

  file_close(0);
}

void handle_wildcard() {
  static char *s;

  filename[0]='$';
  file_open(filename,0); /* Chan 0 = load PRG */
 
  SET_ST(0);
  /* Make device talk the directory. */
  fast1541_disable_screen();
  cbm_k_talk(device);
  cbm_k_tksa(0x60);  /* Chan 0 = load PRG */
  /* Skip load address, BASIC line pointer and line number (file size)*/
  serial_skip(6);
  do {} while (cbm_k_acptr()!=0);
  /* Skip BASIC line pointer and line number (file size)*/
  serial_skip(4);
  do {} while (cbm_k_acptr()!='"');
  s=filename+1;
  do {} while ((*s++=cbm_k_acptr())!='"');
  s--;
  filename_extra=s;
  *s=0;
  cbm_k_untlk();
  fast1541_enable_screen();

  file_close(0);
}

void set_filename(register const char *s) {
  register char *t;
  static char wildcard;

  wildcard=0;
  t=filename+1;
  while ((*t=*s)!=13) {
    if (*s=='*')
      wildcard=1;
    ++s;
    ++t;
  }
  filename_extra=t;
  *t=0;

  if (wildcard)
    handle_wildcard();

  strout("Setting filename to \"");
  strout(filename+1);
  strout("\"\n");
}

//#define cursor(b) {if (b) POKE(204,0); else POKE(204,255);}

extern void swap_font_basicstub();
extern void restore_basicstub();

void dump_tool_to_file() {
  register unsigned s;
  static unsigned t;

  SET_ST(0);

  file_open(filename+1,1);   /* Chan 1 = save PRG */
  if (GET_ST())
    goto error;

  /* Give listen command */
  cbm_k_listen(device);
  cbm_k_second(0x61);  /* Chan 1 = save PRG */
  if (GET_ST())
    goto error;

  /* Write load address 0x0801 */
  cbm_k_ciout(0x01);
  cbm_k_ciout(0x08);

  /* Write BASIC stub */
  s=0x4781;
  t=0x4800;
  do  {
    cbm_k_ciout(PEEK(s));
    ++s;
  } while (s!=t);
  if (GET_ST())
    goto error;

  /* Write high part of font and main program*/
  s=0x0880;
  t=tool_start+tool_size;
  do  {
    cbm_k_ciout(PEEK(s));
    ++s;
  } while (s!=t);
  if (GET_ST())
    goto error;
  
  /* Write fill bytes */
  t=0x4800;
  do  {
    cbm_k_ciout(0xff);
    ++s;
  } while (s!=t);
  if (GET_ST())
    goto error;

  /* Write low part of font */
  s=0x0800;
  t=0x0880;
  do  {
    cbm_k_ciout(PEEK(s));
    ++s;
  } while (s!=t);
  if (GET_ST())
    goto error;

  cbm_k_unlsn();
  strout("\x1e" "File written." "\x9a\n");
  goto end;

error:
  handle_st(GET_ST());
//  strout("Failed to write file.\n");
end:
  file_close(1);
  filename_extra[0]=0;
}

char test_roml_ram() {
  POKE(0x8000,0);
  if (PEEK(0x8000)!=0)
    return 0;
  POKE(0x8000,0xff);
  if (PEEK(0x8000)!=0xff)
    return 0;
  return 1;
}

char test_romh_ram() {
  POKE(0xa000,0);
  if (PEEK(0xa000)!=0)
    return 0;
  POKE(0xa000,0xff);
  if (PEEK(0xa000)!=0xff)
    return 0;
  return 1;
}

#define green_checkmark "\x1e\xba\x9a\n"
#define red_minus "\x1c-\x9a\n"

char fc3_detect() {
  /* Returns:
     0 - No cart (ROML is RAM)
     1 - Gmod2
     2 - Magic desk
     3 - Unknown cart (ROML stuck at ROM)
   */
  static char c;

  strout("Testing cartridge:\n");
  fc3_detected=1;

  strout("EXROM control    ");
  FC3_BANK(0x60);  /* Activate 8K cartridge mode */
  c=test_roml_ram();
  FC3_BANK(0x70);  /* Activate normal mode */
  if (c) {
    strout(red_minus);
    fc3_detected=0;
  } else
    strout(green_checkmark);

  strout("GAME control     ");
  if (test_ultimax())
    strout(green_checkmark);
  else {
    strout(red_minus);
    fc3_detected=0;
  }

  strout("NMI              ");
  if (test_nmi())
    strout(green_checkmark);
  else {
    strout(red_minus);
    fc3_detected=0;
  }

  cbm_k_bsout('\n');
  return 0;
}

void flash_mode_autoselect_5v() {
  /* put the flash ROM in its "autoselect" mode */
  flash_magic_unlock();
  FLASH_POKE(1,0x9555,0x90);
}

void flash_mode_reset_5v() {
  FLASH_POKE(1,0x9555,0xf0);
}

void flash_mode_autoselect_12v() {
  /* put the flash ROM in its "autoselect" mode */
  FLASH_POKE(1,0x9555,0x90);
}

void flash_mode_reset_12v() {
  /* Return to read mode */
  FLASH_POKE(1,0x9555,0x00);
}


/*
char flash_write_byte_5v(const char bank, unsigned offset, const char c) {
  static char count;

  offset|=0x8000;
  flash_magic_unlock();
  FLASH_POKE(1,0x9555,0xa0);
  FLASH_POKE(bank,offset,c);
  count=192;
  while (PEEK(offset)!=c) {
    if (++count == 0)
      return 0;
  }
  return 1;
}
*/

#define flash_write_byte_5v(bank,offset,c) ( \
  __A__ = (c),								\
  asm("sta tmp1"),							\
  __A__ = (bank),							\
  asm("ora #$50"),							\
  asm("sta tmp2"),							\
  __AX__ = (offset),						\
  asm("sta ptr1"),							\
  asm("txa"),								\
  asm("ora #$80"),							\
  asm("sta ptr1+1"),						\
  asm("jsr asm_flash_write_byte_5v"),		\
  __A__ \
)

/*
char flash_write_byte_12v(const char bank, const unsigned offset, const char c) {
  static char count;

  count=25;
  do {
    FLASH_POKE(bank,offset,0x40);
    FLASH_POKE(bank,offset,c);
    FLASH_POKE(bank,offset,0xc0);
    if (PEEK(0x8000|offset)==c)
      return 1;
  } while (--count>0);
  return 0;
}
*/

#define flash_write_byte_12v(bank,offset,c) ( \
  __A__ = (c),								\
  asm("sta tmp1"),							\
  __A__ = (bank),							\
  asm("sta tmp2"),							\
  __AX__ = (offset),						\
  asm("sta ptr1"),							\
  asm("stx ptr1+1"),						\
  asm("jsr asm_flash_write_byte_12v"),		\
  __A__ \
)


char flash_erase_command_5v(char bank,char offset) {
  static char count;
  static char v;

  /* Critical timing, delay of interrupt in the middle unacceptable */
  SEI();
  flash_magic_unlock();
  FLASH_POKE(1,0x9555,0x80);
  flash_magic_unlock();
  FLASH_POKE(bank,offset<<8,0x30)
  CLI();
  /* Wait at least 8 seconds */
  count=252;
retry:
  /* TI clock middle byte changes every 255/60 = 4.25 sec on 60Hz */
  v = PEEK(TI_MID);

  do {
    if (PEEK(TI_MID)!=v) {
      if (++count==0)  /* 60 Hz = worst case */
        return 0;  /* TImeout */
      goto retry;
    }
  } while (PEEK(offset<<8)!=255);
  return 1;
}

char flash_erase_sector_5v(char sector) {
  static char offset,bank,highbank;

  strout("Erasing sector ");
  numout(sector);
  cbm_k_bsout('\n');
  if (flashsecsz==12) {
    /* 12 */
    offset=(sector & 3) << 4;
    bank=sector>>2;
    highbank=0;
  } else if (flashsecsz==16) {
    /* 16 */
    offset=0;
    bank=sector<<2;
    highbank=bank;
  } else if (flashsecsz==17) {
    /* 17 */
    offset=0;
    bank=sector<<3;
    highbank=bank|4;
  } else {
    /* 18 */
    offset=0;
    bank=0;
    highbank=0;
  }
  offset|=0x80;
  if ((sector==0) && (flash_features[flashidx] & flash_bottomboot)) {
    if (!flash_erase_command_5v(bank|1,offset))
      return 0;
    if (!flash_erase_command_5v(bank|1,offset|0x20))
      return 0;
    if (!flash_erase_command_5v(bank|2,offset))
      return 0;
  }
  if ((sector==flashseccnt-1) && (flash_features[flashidx] & flash_topboot)) {
    if (!flash_erase_command_5v(highbank|2,offset))
      return 0;
    if (!flash_erase_command_5v(highbank|2,offset|0x20))
      return 0;
    if (!flash_erase_command_5v(highbank|3,offset))
      return 0;
  }
  return flash_erase_command_5v(bank,offset);
}

extern char flash_sector_is_empty_12v(register char sector);
/*char flash_sector_is_empty_12v(register char sector) {
  static char banks,bank,b;
  static unsigned o;

  banks=1<<(flashsecsz-14);
  bank=sector<<(flashsecsz-14);
  for (b=bank;b<bank+banks;b++) {
    FC3_BANK(0x40|b);
    for (o=0x8000;o<0xc000;o++)
      if (PEEK(o)!=0xff)
        return 0;
  }
  return 1;
}*/

char flash_clear_sector_12v(char sector) {
  static char ebank,b;
  static unsigned o;

  strout("Clearing sector ");
  numout(sector);
  cbm_k_bsout('\n');

  b=sector<<(flashsecsz-14);
  ebank=b + 1<<(flashsecsz-14);
  do {
    for (o=0x8000;o<0xc000;++o)
      if (flash_write_byte_12v(b,o,0)==0)
        return 0;
  } while (++b!=ebank);
  return 1;
}

char flash_erase_sector_12v(char sector) {
  static char ebank,b;
  static unsigned i,erases;
  register unsigned o;

  strout("Erasing sector ");
  numout(sector);
  cbm_k_bsout('\n');

  b=sector<<(flashsecsz-14);
  ebank=b + 1<<(flashsecsz-14);
  erases=1000;
  do {
    for (o=0x8000;o<0xc000;o++) {
      while (PEEK(o)!=0xff) {
        if (--erases == 0)
          return 0;
        FLASH_POKE(b,o,0x20);
        FLASH_POKE(b,o,0x20);
        for (i=2000;i!=0;i--) {};
        FLASH_POKE(b,o,0xa0);
      }
    }
  } while (++b!=ebank);
  return 1;
}


void write_tool_to_flash() {
  register unsigned s = 0;

  register char i,z;


  if FLASHIDX_IS_12V(flashidx) {
    flash_enable_12v();
    if (flash_sector_is_empty_12v(0)) {
      strout("Sector 0 is empty, skipping erase.\n");
    } else {
      if (!flash_clear_sector_12v(0))
        goto clear_error;
      if (!flash_erase_sector_12v(0))
        goto erase_error;
    }
    strout("Writing tool\n");
    do {
      if (!flash_write_byte_12v(0,s|0x8000,PEEK(tool_start+s)))
        goto error;
    } while (++s!=tool_size);
    for (s=0x3780;s<0x3800;s++)
      if (!flash_write_byte_12v(0,s|0x8000,PEEK(tool_start+s)))
        goto error;
    for (s=0x3800;s<0x4000;s++)
      if (!flash_write_byte_12v(0,s|0x8000,PEEK(s-0x3000)))
        goto error;
  } else {
    z=1;
    if (flashsecsz<=12)
      z=4;
    if (secprot[0] && (flash_features[flashidx] & flash_12v_unprotect)) {
      strout("Enabling 12V to override write-protect.\n");
      flash_enable_12v();
    }
    for (i=0;i<z;i++)
      if (!flash_erase_sector_5v(0))
        goto erase_error;
    strout("Writing tool\n");
    do {
      if (!flash_write_byte_5v(0,s,PEEK(tool_start+s)))
        goto error;
    } while (++s!=tool_size);
    for (s=0x3780;s<0x3800;s++)
      if (!flash_write_byte_5v(0,s,PEEK(tool_start+s)))
        goto error;
    for (s=0x3800;s<0x4000;s++)
      if (!flash_write_byte_5v(0,s,PEEK(s-0x3000)))
        goto error;
  }
  return;
clear_error:
  strout("Flash clear error in sector 0");
  goto end;
erase_error:
  strout("Flash erase error in sector 0");
  goto end;
error:
  strout("Flash write error at ");
  numout(s);
  cbm_k_bsout('\n');
end:
  flash_disable_12v();
}


void flash_detect() {
   static unsigned flashid,p,cnt,old_m,old_d,valid_flash_id,test_12v;
   register char i;

  test_12v=0;
  old_m=PEEK(0x8000);
  old_d=PEEK(0x8001);
  flash_mode_autoselect_5v();
  FC3_BANK(0x40);
retest:
  valid_flash_id=(PEEK(0x8000)!=old_m) || (PEEK(0x8001)!=old_d);

  flashid=(PEEK(0x8000)<<8) | PEEK(0x8001);
  cbm_k_bsout(13);
  for (i=0x43; i&0x40 ;i--) {
    if (test_12v) {
      flash_mode_reset_12v();
      flash_mode_autoselect_12v();
    } else {
      flash_mode_reset_5v();
      flash_mode_autoselect_5v();
    }
    FC3_BANK(i);  /* Swap banks to reduce chance of collision with flash contents */
    if (flashid!=((PEEK(0x8000)<<8) | PEEK(0x8001))) {
      cbm_k_bsout(13);
      valid_flash_id=0;
      break;
    }
  };
  if (!valid_flash_id)
    if (!test_12v) {
      flash_mode_reset_5v();
      test_12v++;
      flash_enable_12v();
      strout("Did not find 5V flash ROM.\nSwitched on 12V generator, retrying.\n");
      flash_mode_autoselect_12v();
      goto retest;
    } else {
      flash_mode_reset_12v();
      flash_disable_12v();
      strout("There appears to be no flash ROM on the cartridge.\n");
      return;
    };
  if (flashid==0x7f7f)
    /* EON flash ROMS need this */
    flashid=(PEEK(0x8100)<<8) | PEEK(0x8101);
//  strout("Flash ID: ");
//  numout(flashid);
//  cbm_k_bsout('\n');
  flashidx=0xff;
  cnt=0;
  for (i=0;i<flash_chips_count;i++)
    if (flashrom_ids[i]==flashid) {
      flashidx=i;
      cnt=seccnt[flashidx];
      break;
    }
  for (i=0;i<8;i++)
    secprot[i]=0;
  if (flashidx!=0xff) {
    strout("\x1e" "Found" "\x9a ");
    strout(flashrom_names[flashidx]);
    if (flash_features[flashidx] & flash_sectorprotect) {
      strout(", writeable sectors ");
      for (i=0;i<cnt;i++) {
        FC3_BANK(i<<2 | 0x40);
        p=PEEK(0x8002);
        if ((i==0 && flash_features[flashidx] & flash_bottomboot) ||
            (i==cnt-1 && flash_features[flashidx] & flash_topboot)) {
          FC3_BANK(i<<2 | 0x41);
          p|=PEEK(0x8002);
          p|=PEEK(0xa002);
          FC3_BANK(i<<2 | 0x42);
          p|=PEEK(0x8002);
        }
        if (p)
          cbm_k_bsout('-');
        else
          cbm_k_bsout(186);  /* Checkmark*/
        secprot[i]=p;
      };
    }
    flashsecsz=secsz[flashidx];
    cbm_k_bsout('\n');
    strout("Sector cnt: ");
    numout(cnt);
    strout(" size: 2^");
    numout(flashsecsz);
    cbm_k_bsout('\n');

    if (flashseccnt!=seccnt[flashidx]) {
      flashseccnt=cnt;
      strout("seca=0 secz=");
      seca=0;
      secz=flashseccnt-1;
      numout(secz);
    }
  } else {
    strout("A flash might be present, but it's ID\nwasn't recognized.");
  }
  cbm_k_bsout('\n');
  if (test_12v) {
    flash_mode_reset_12v();
    flash_disable_12v();
  } else
    flash_mode_reset_5v();
};

void filename_set_extra(char c) {
  filename_extra[0]=',';
  filename_extra[1]='s';
  filename_extra[2]=',';
  filename_extra[3]=c;
  filename_extra[4]=0;
}

void flash_write_file() {
  register unsigned cursec,secptr,secend;
  static unsigned base;
  static char ebank,curbank,b;

  filename_set_extra('r');
  SET_ST(0);
  file_open(filename+1,2);
  if (GET_ST())
    goto file_error;
  fast1541_disable_screen();
  cbm_k_talk(device);
  cbm_k_tksa(0x62);
  if (GET_ST())
    goto file_error;
  for (cursec=seca;cursec<=secz;cursec++) {
    if FLASHIDX_IS_12V(flashidx) {
      flash_enable_12v();
      if (flash_sector_is_empty_12v(cursec)) {
        strout("Sector ");
        numout(cursec);
        strout(" is empty, skipping erase.\n");
      } else {
        if (!flash_clear_sector_12v(cursec))
          goto clear_error;
        if (!flash_erase_sector_12v(cursec))
          goto erase_error;
      }
    } else {
      if (secprot[cursec] && (flash_features[flashidx] & flash_12v_unprotect)) {
        strout("Enabling 12V to override write-protect.\n");
        flash_enable_12v();
      }
      if (!flash_erase_sector_5v(cursec))
        goto erase_error;
    }
    strout("Writing from file to flash sector ");
    numout(cursec);
    cbm_k_bsout('\n');
    if FLASHIDX_IS_12V(flashidx) {
      curbank=cursec<<(flashsecsz-14);
      ebank=curbank + 1<<(flashsecsz-14);
      do {
        for (secptr=0x8000;secptr<0xc000;secptr++) {
          b=cbm_k_acptr();
          if (GET_ST() && (secptr!=0xbfff)) {
//            secptr&=0x7fff;
            goto file_error;
          }
          if (!flash_write_byte_12v(curbank,secptr,b)) {
            secptr&=0x7fff;
            goto flash_error;
          }
        }
      } while (++curbank!=ebank);
    } else {
      if (flashsecsz>=17) {
        curbank=cursec<<(flashsecsz-14);
        ebank=curbank + 1<<(flashsecsz-14);
        do
          for (secptr=0x0000;secptr<0x4000;secptr++) {
            b=cbm_k_acptr();
            if (GET_ST() && (secptr!=0x3fff))
              goto file_error;
            if (!flash_write_byte_5v(curbank,secptr,b))
              goto flash_error;
          }
        while (++curbank!=ebank);
      } else {
        /* flashsecsz<17 */
        secptr=0;
        base=0;
        if (flashsecsz==12) {
          base=(cursec & 3) << 12;
          curbank=cursec>>2;
          secend=0x1000;
        } else {
          /* 16 */
          secend=0;
          curbank=(cursec<<2) | (secptr>>14);
        }
        do {
          if ((secptr & 0x3fff) == 0) {
            if (flashsecsz==12) {
              curbank=cursec>>2;
            } else {
              curbank=(cursec<<2) | (secptr>>14);
            }
          }
          b=cbm_k_acptr();
          if (GET_ST() && (secptr!=(secend-1)))
            goto file_error;
          if (!flash_write_byte_5v(curbank,base|(secptr & 0x3fff),b))
            goto flash_error;
          ++secptr;
        } while (secptr!=secend);
      }
    }
  }
  
  strout("\x1e" "Flash write succesfully completed.\x9a\n");
  goto end;
file_error:
  if ((GET_ST() & 64) && ((secptr & 0x3fff)==0))
    strout("End of file at bank boundary, this\nlikely was intended and the operation\nwas succesfull.\n");
  else
    handle_st(GET_ST());
  goto end;
clear_error:
  strout("Flash clear error in sector ");
  goto out_cursec;
erase_error:
  strout("Flash erase error in sector ");
out_cursec:
  numout(cursec);
  cbm_k_bsout('\n');
  goto end;
flash_error:
  strout("Flash write error at ");
  numout(curbank);
  cbm_k_bsout(':');
  numout(secptr);
  cbm_k_bsout('\n');
end:
  flash_disable_12v();
  cbm_k_untlk();
  fast1541_enable_screen();
  filename_extra[0]=0;
  file_close(2);
}


void flash_dump_to_file() {
  register unsigned cursec,secptr,secend;
  static char bank,ebank;

  if (!fc3_detected) {
    strout("No cart error.\n");
    return;
  }
  filename_set_extra('w');
  file_open(filename+1,2);
  cbm_k_listen(device);
  cbm_k_second(0x62);
  for (cursec=seca;cursec<=secz;cursec++) {
    strout("Writing flash sector ");
    numout(cursec);
    strout(" to file.\n");
    if (flashsecsz>=17) {
      bank=cursec<<(flashsecsz-14);
      ebank=bank + 1<<(flashsecsz-14);
      do {
        FC3_BANK(0x40|bank);
        for (secptr=0x8000;secptr<0xc000;secptr++) {
          cbm_k_ciout(PEEK(secptr));
        }
      } while (++bank!=ebank);
    } else {
      /* flashsecsz<17 */
      secptr=0;
      if (flashsecsz==12) {
        secptr=(cursec & 3) << 12;
        bank=cursec>>2;
        secend=0x1000;
      } else {
        /* 16 */
        secend=0;
        bank=(cursec<<2) | (secptr>>14);
      }
      FC3_BANK(0x40|bank);
      do {
        cbm_k_ciout(PEEK(0x8000 | (secptr & 0x3fff)));
        ++secptr;
        if ((secptr & 0x3fff) == 0) {
          if (flashsecsz==16) {
            bank++;
            FC3_BANK(0x40|bank);
          }
        }
      } while (secptr!=secend);
    }
  }
  cbm_k_unlsn();
  filename_extra[0]=0;
  file_close(2);
}


void print_header() {
  static const char header[] = {
    p_clr,
    p_hlb,p_hor,p_hor,p_hor,p_hor,p_hor,p_hor,p_hor,p_hor,p_hor,p_hor,p_hor,p_hor,p_hor,p_hor,p_hor,p_hor,p_hor,p_hor,p_hor,p_hor,p_hor,p_hor,p_hor,p_hor,p_hor,p_hrb,  13,
    p_ver,'F'  ,'i',  'n',  'a',  'l',  ' ',  'C',  'a',  'r',  't',  'r',  'i',  'd'  ,'g'  ,'e'  ,' '  ,'3'  ,' '  ,'f'  ,'l'  ,'a'  ,'s'  ,'h'  ,'e'  ,'r'  ,p_ver,  13,
    p_hlo,p_hor,p_hor,p_hor,p_hor,p_hor,p_hor,p_hor,p_hor,p_hor,p_hor,p_hor,p_hor,p_hor,p_hor,p_hor,p_hor,p_hor,p_hor,p_hor,p_hor,p_hor,p_hor,p_hor,p_hor,p_hor,p_hro,  13,
    13,0
  };

  strout(header);
  strout("(C) 2023 by Daniel Mantione\n\n");
}

void print_help() {
  strout("Commands you enter are passed to the de-"
         "vice as DOS commands. To switch between\n"
         "devices, enter their device number.\n\n");
  strout("Special commands:\n"
          "* = wildcard supported\n"
          "help    - This help text.\n"
          "$       - directory*\n"
          "file:fn - Set \"fn\" as file to flash*\n"
          "seca:n  - \"n\" = first sector to flash\n"
          "secz:n  - \"n\" = last sector to flash\n"
          "fdump   - Dump flash to file\n"
          "tdump   - Dump this tool to file\n");
  if (flashidx!=0xff) {
    strout("fwrit   - Write file to flash\n"
           "twrit   - Write this tool to flash\n");
  } else {
    strout("won     - Enable write commands\n");
  }
  if(started_from_basic)
    strout("exit    - Exit to BASIC\n");
  else
    strout("exit    - Boot to BASIC\n");
}

void input_command() {
  register char *s;
  numout(device & 0x7f);
  if (device & 0x80)
    strout("!!");
  strout("> ");
  s=tmpstr;
  do {
    *s=cbm_k_basin();
  } while (*s++!=13);
  *s=0;
}

void main_loop() {
  register char *s,*t;
  do {
    if (device!=olddevice) {
      char c;
      c=send_drive_command("ui");
      handle_st(c);
      device|=c & 0x80;
      olddevice=device;
    }
    if ((device & 0x80) == 0)
      print_drive_status();
    input_command();
    cbm_k_bsout(13);
    s=tmpstr;
    while (*s==' ')
      s++;
    /* Directory */
    if (s[0]=='$' && (device & 0x80) == 0) {
      t=s;
      while (*t!=13)
        ++t;
      *t=0;
      directory(s);
    }
    /* Device ID*/
    else if (s[0]>='0' && s[0]<='9' && s[1]==13 && !fast1541)
      device=s[0]-'0';
    else if (((s[0]>='0' && s[0]<='2' && s[1]>='0' && s[1]<='9') || (s[0]=='3' && s[1]=='0')) && s[2]==13 && !fast1541)
      device=10*(s[0]-'0')+(s[1]-'0');
    /* Exit */
    else if (s[0]=='e') { 
      if (s[1]=='x' && s[2]=='i' && s[3]=='t')
        break;
    } else if (s[0]=='f' && (device & 0x80) == 0) {
      /* fdump */
      if (s[1]=='d' && s[2] =='u' && s[3]=='m' && s[4]=='p')
        flash_dump_to_file();
      /* File */
      if (s[1]=='i' && s[2] =='l' && s[3]=='e' && s[4]==':')
        set_filename(s+5);
      /* fwrit */
      if (s[1]=='w' && s[2] =='r' && s[3]=='i' && s[4]=='t' && flashidx!=0xff)
        flash_write_file();
    }
    /* Help */
    else if (s[0]=='h' && s[1]=='e' && s[2]=='l' && s[3]=='p')
      print_help();
    else if (s[0]=='s' && s[1]=='e' && s[2]=='c' && s[4]==':') {
      if (s[3]=='a') {
        /* seca */
        seca=atoi(s+5);
        strout("Start");
      } else {
        /* secz */
        secz=atoi(s+5);
        strout("End");
      };
      strout(" sector set to ");
      strout(s+5);
    } else if (s[0]=='t' && (device & 0x80) == 0) {
      /* Dump tool to file */
      if (s[1]=='d' && s[2] =='u' && s[3]=='m' && s[4]=='p')
        dump_tool_to_file();
      if (s[1]=='w' && s[2] =='r' && s[3]=='i' && s[4]=='t' && flashidx!=0xff)
        write_tool_to_flash();
    } else if (s[0]=='w' && s[1]=='o' && s[2]=='n') {
      flash_detect();
    } else
      if (device & 0x80)
        handle_st(device & 0x80);
      else
        send_drive_command(s);
      if (s[0]=='u' && s[1]=='i') {
        if (s[2]=='-')
          fast1541=1;
        if (s[2]=='+')
          fast1541=0;
      }
  } while (1);
}

int main (void) {
  /* Enable ROM */
  POKE(1,7);
  started_from_basic=PEEK(0x0800)==0;
  install_nmi();
  if (started_from_basic)
    /*Swap BASIC stub and font */
    swap_font_basicstub();
  POKE(0xd018,0x12);

  fast1541=0;
  olddevice=0;
  device=8;
  flashidx=255;
  seca=0;
  secz=3;
  flashseccnt=4;

  install_lowcode();

  print_header();
  fc3_detect();
  strout("\nType \"help\" for help.\n");

  main_loop();

  /* Reset cartridge. */
  FC3_BANK(0x70);

  strout("load \"flasher-c64.prg\",10\n");

  POKE(0xd018,0x14);
  if (started_from_basic)
    /*Swap BASIC stub and font */
    restore_basicstub();
  return EXIT_SUCCESS;
}
