
.include "cbm_kernal.inc"

.segment "CODE"

.import __FONTL_START__,__BASICSTUBH_START__
.import __LOWCODE1_LOAD__, __LOWCODE1_RUN__, __LOWCODE1_SIZE__
.import __LOWCODE2_LOAD__, __LOWCODE2_RUN__, __LOWCODE2_SIZE__
.importzp tmp1,tmp2,ptr1,ptr2
.importzp _flashsecsz

.export _swap_font_basicstub
.proc _swap_font_basicstub: near

		ldx #$7f
L1:
		lda $0800,x
		sta __BASICSTUBH_START__,x
		lda __FONTL_START__,x
		sta $0800,x
		dex
		bpl L1
		rts
.endproc

.export _restore_basicstub
.proc _restore_basicstub
		ldx #$7f
L1:
		lda __BASICSTUBH_START__,x
		sta $0800,x
		dex
		bpl L1
		rts
.endproc

.export _strout
.proc _strout: near

		sta L1+1
		stx L1+2
L1:
		lda $ffff ; address overwritten
		beq L2
		jsr BSOUT
		inc L1+1
		bne L1
		inc L1+2
		bne L1
L2:
		rts

.endproc

.export _serialout
.proc _serialout: near

		sta L1+1
		stx L1+2
L1:
		lda $ffff ; address overwritten
		beq L2
		jsr CIOUT
		inc L1+1
		bne L1
		inc L1+2
		bne L1
L2:
		rts

.endproc

.export _serial_skip
.proc _serial_skip

		tax
L1:
		jsr ACPTR
		dex
		bne L1
		rts

.endproc

.proc nmi_handler
		lda #1
		rti
		.word $fff7
		.word $fff8
		.word $fff8
.endproc


.export _install_nmi
.proc _install_nmi
		ldx #.SIZEOF(nmi_handler)-1
L1:
		lda nmi_handler,x
		sta $fff7,x
		dex
		bpl L1
		rts
.endproc

.export _test_nmi
.proc _test_nmi
		sei
		ldx #5
		stx $1
		ldx #$30
		stx $dfff  ; Generate NMI (sets A to 1)
		lda #0
		ldx #$70
		stx $dfff  ; Clear NMI
		ldx #7
		stx $1
		cli
		rts
.endproc

.export _install_lowcode
.proc _install_lowcode
		ldx #<__LOWCODE1_SIZE__
:		lda __LOWCODE1_LOAD__-1,x
		sta __LOWCODE1_RUN__-1,x
		dex
		bne :-
		ldx #<__LOWCODE2_SIZE__
:		lda __LOWCODE2_LOAD__-1,x
		sta __LOWCODE2_RUN__-1,x
		dex
		bne :-
		rts
.endproc

.export _flash_enable_12v
.proc _flash_enable_12v
		stx $dffe
		stx $dffe
		stx $dffe
		stx $dffe
		rts
.endproc

.export _flash_disable_12v
.proc _flash_disable_12v
		stx $df00
		rts
.endproc

;
; Speed optimized routine to check if flash is empty, scans 256KB in 3s.
;

.export _flash_sector_is_empty_12v
.proc _flash_sector_is_empty_12v
		sta tmp1

		lda _flashsecsz
		sec
		sbc #14
		pha
	
		tay
		beq @3
		lda #1
@1:		asl a
		dey
		bne @1
@3:		sta tmp2 ; Number of banks to check

		pla
		tay
		beq @4
		lda tmp1
@2:		asl a
		dey
		bne @2
@4:		clc
		sta tmp1 ; Start bank
		adc tmp2
		sta tmp2 ; End bank

		lda tmp1
@nb:	pha
		ora #$40
		sta $dfff

		lda #$80
		sta @and+2

@tp:	ldy #0
		lda #$ff
@and:	and $8000,y
		iny
		bne @and
		cmp #$ff
		bne @notzero
		inc @and+2
		lda @and+2
		cmp #$c0
		bcc @tp
		; C is set
		pla
		adc #0 ; C still set
		cmp tmp2
		bcc @nb

		lda #1
		rts
@notzero:
		pla
		lda #0
		rts
.endproc

.SEGMENT "LOWCODE1"

.export _test_ultimax
.proc _test_ultimax
		lda #1
		sei
		; Enter Ultimax mode
		ldx #$50
		stx $dfff
		ldy #4
@2:		; Try to write to deactivated RAM
		ldx #$aa
		stx $7fff
		cpx $7fff
		bne @1
		; So far so good, try another value
		ldx #$55
		stx $7fff
		cpx $7fff
		bne @1
		dey
		bne @2
		lda #0
@1:		ldx #$40
		stx $dfff
		cli
		rts
.endproc

.SEGMENT "LOWCODE2"


.export _flash_magic_unlock
.proc _flash_magic_unlock
		sei
		; Enter Ultimax mode in bank 1
		ldx #$51
		stx $dfff
		; First step of the magical unlock dance
		lda #$aa
		sta $9555
		; Ultimax mode in bank 0
		ldx #$50
		stx $dfff
		; Second step of the magical unlock dance
		lsr a
		sta $eaaa
		; Exit Ultimax mode
		ldx #$40
		stx $dfff
		cli
		rts
.endproc

.export _flash_poke
.proc _flash_poke
		; Poke a byte to flash memory
		sei
.export _pokebank = * + 1
		ldx #$50		; Modified before call
		stx $dfff
.export _pokeaddr = * + 1
		sta $8000		; Modified before call
		lda _pokebank
		and #$cf
		sta $dfff
		cli
		rts
.endproc

.export asm_flash_write_byte_5v
.proc asm_flash_write_byte_5v
		sei
		ldx #$51		; Bank 1 in Ultimax mode
		stx $dfff

		; First step of the magical unlock dance
		lda #$aa
		sta $9555
		; Ultimax mode in bank 0
		ldy #$50
		sty $dfff
		; Second step of the magical unlock dance
		lsr a
		sta $eaaa

		stx $dfff
		lda #$a0		; Flash write byte command
		sta $9555

		lda ptr1+1
		and #$20
		asl
		ora ptr1+1
		sta ptr1+1

		ldx tmp2		; Bank to write byte to
		stx $dfff
		lda tmp1		; Byte to write
		ldy #0
		sta (ptr1),y	; Write the byte
		ldx #$c0
@wait:	cmp (ptr1),y	; Wait until correct value is read
		beq @ok			; Note: C=1 when equal
		inx
		bne @wait
		clc
@ok:	lda #0			; Success
		rol a			; C=1 if jumped to @ok
		ldx #$40
		stx $dfff
		cli
		rts
.endproc

.export asm_flash_write_byte_12v
.proc asm_flash_write_byte_12v
		sei
		lda tmp2 ; bank
		ora #$50 ; Ultimax
		sta $dfff

		lda ptr1+1
		and #$20
		asl
		ora ptr1+1
		sta ptr1+1

		ldx #25
@retry: lda #$40
		ldy #0
		sta (ptr1),y
		lda tmp1 ; data
		sta (ptr1),y
		nop		; delay
		nop		; delay
		lda #$c0
		sta (ptr1),y
		nop		; delay
		lda tmp1 ; data
		cmp (ptr1),y	; Also sets C=1 if equal
		beq @ok
		dex
		bne @retry
		clc
@ok:	lda tmp2
		ora #$40
		sta $dfff
		lda #0
		rol a
		cli
		rts
.endproc
